import unittest

import reactivex
from reactivex import operators as ops
from reactivex.testing import ReactiveTest, TestScheduler

on_next = ReactiveTest.on_next
on_completed = ReactiveTest.on_completed
on_error = ReactiveTest.on_error
subscribe = ReactiveTest.subscribe
subscribed = ReactiveTest.subscribed
disposed = ReactiveTest.disposed
created = ReactiveTest.created


class TestMerge(unittest.TestCase):
    def test_merge_never2(self):
        scheduler = TestScheduler()
        n1 = reactivex.never()
        n2 = reactivex.never()

        def create():
            return reactivex.merge(n1, n2)

        results = scheduler.start(create)
        assert results.messages == []

    def test_merge_never3(self):
        scheduler = TestScheduler()
        n1 = reactivex.never()
        n2 = reactivex.never()
        n3 = reactivex.never()

        def create():
            return reactivex.merge(n1, n2, n3)

        results = scheduler.start(create)
        assert results.messages == []

    def test_merge_empty2(self):
        scheduler = TestScheduler()
        e1 = reactivex.empty()
        e2 = reactivex.empty()

        def create():
            return reactivex.merge(e1, e2)

        results = scheduler.start(create)
        assert results.messages == [on_completed(200)]

    def test_merge_empty3(self):
        scheduler = TestScheduler()
        e1 = reactivex.empty()
        e2 = reactivex.empty()
        e3 = reactivex.empty()

        def create():
            return reactivex.merge(e1, e2, e3)

        results = scheduler.start(create)
        assert results.messages == [on_completed(200)]

    def test_merge_empty_delayed2_right_last(self):
        scheduler = TestScheduler()
        l_msgs = [on_next(150, 1), on_completed(240)]
        r_msgs = [on_next(150, 1), on_completed(250)]
        e1 = scheduler.create_hot_observable(l_msgs)
        e2 = scheduler.create_hot_observable(r_msgs)

        def create():
            return reactivex.merge(e1, e2)

        results = scheduler.start(create)
        assert results.messages == [on_completed(250)]

    def test_merge_empty_delayed2_left_last(self):
        scheduler = TestScheduler()
        l_msgs = [on_next(150, 1), on_completed(250)]
        r_msgs = [on_next(150, 1), on_completed(240)]
        e1 = scheduler.create_hot_observable(l_msgs)
        e2 = scheduler.create_hot_observable(r_msgs)

        def create():
            return reactivex.merge(e1, e2)

        results = scheduler.start(create)
        assert results.messages == [on_completed(250)]

    def test_merge_empty_delayed3_middle_last(self):
        scheduler = TestScheduler()
        msgs1 = [on_next(150, 1), on_completed(245)]
        msgs2 = [on_next(150, 1), on_completed(250)]
        msgs3 = [on_next(150, 1), on_completed(240)]
        e1 = scheduler.create_hot_observable(msgs1)
        e2 = scheduler.create_hot_observable(msgs2)
        e3 = scheduler.create_hot_observable(msgs3)

        def create():
            return reactivex.merge(e1, e2, e3)

        results = scheduler.start(create)
        assert results.messages == [on_completed(250)]

    def test_merge_empty_never(self):
        scheduler = TestScheduler()
        msgs1 = [on_next(150, 1), on_completed(245)]
        e1 = scheduler.create_hot_observable(msgs1)
        n1 = reactivex.never()

        def create():
            return reactivex.merge(e1, n1)

        results = scheduler.start(create)
        assert results.messages == []

    def test_merge_never_empty(self):
        scheduler = TestScheduler()
        msgs1 = [on_next(150, 1), on_completed(245)]
        e1 = scheduler.create_hot_observable(msgs1)
        n1 = reactivex.never()

        def create():
            return reactivex.merge(n1, e1)

        results = scheduler.start(create)
        assert results.messages == []

    def test_merge_return_never(self):
        scheduler = TestScheduler()
        msgs1 = [on_next(150, 1), on_next(210, 2), on_completed(245)]
        r1 = scheduler.create_hot_observable(msgs1)
        n1 = reactivex.never()

        def create():
            return reactivex.merge(r1, n1)

        results = scheduler.start(create)
        assert results.messages == [on_next(210, 2)]

    def test_merge_never_return(self):
        scheduler = TestScheduler()
        msgs1 = [on_next(150, 1), on_next(210, 2), on_completed(245)]
        r1 = scheduler.create_hot_observable(msgs1)
        n1 = reactivex.never()

        def create():
            return reactivex.merge(n1, r1)

        results = scheduler.start(create)
        assert results.messages == [on_next(210, 2)]

    def test_merge_error_never(self):
        ex = "ex"
        scheduler = TestScheduler()
        msgs1 = [on_next(150, 1), on_next(210, 2), on_error(245, ex)]
        e1 = scheduler.create_hot_observable(msgs1)
        n1 = reactivex.never()

        def create():
            return reactivex.merge(e1, n1)

        results = scheduler.start(create)
        assert results.messages == [on_next(210, 2), on_error(245, ex)]

    def test_merge_never_error(self):
        ex = "ex"
        scheduler = TestScheduler()
        msgs1 = [on_next(150, 1), on_next(210, 2), on_error(245, ex)]
        e1 = scheduler.create_hot_observable(msgs1)
        n1 = reactivex.never()

        def create():
            return reactivex.merge(n1, e1)

        results = scheduler.start(create)
        assert results.messages == [on_next(210, 2), on_error(245, ex)]

    def test_merge_empty_return(self):
        scheduler = TestScheduler()
        msgs1 = [on_next(150, 1), on_completed(245)]
        msgs2 = [on_next(150, 1), on_next(210, 2), on_completed(250)]
        e1 = scheduler.create_hot_observable(msgs1)
        r1 = scheduler.create_hot_observable(msgs2)

        def create():
            return reactivex.merge(e1, r1)

        results = scheduler.start(create)
        assert results.messages == [on_next(210, 2), on_completed(250)]

    def test_merge_return_empty(self):
        scheduler = TestScheduler()
        msgs1 = [on_next(150, 1), on_completed(245)]
        msgs2 = [on_next(150, 1), on_next(210, 2), on_completed(250)]
        e1 = scheduler.create_hot_observable(msgs1)
        r1 = scheduler.create_hot_observable(msgs2)

        def create():
            return reactivex.merge(r1, e1)

        results = scheduler.start(create)
        assert results.messages == [on_next(210, 2), on_completed(250)]

    def test_merge_lots2(self):
        scheduler = TestScheduler()
        msgs1 = [
            on_next(150, 1),
            on_next(210, 2),
            on_next(220, 4),
            on_next(230, 6),
            on_next(240, 8),
            on_completed(245),
        ]
        msgs2 = [
            on_next(150, 1),
            on_next(215, 3),
            on_next(225, 5),
            on_next(235, 7),
            on_next(245, 9),
            on_completed(250),
        ]
        o1 = scheduler.create_hot_observable(msgs1)
        o2 = scheduler.create_hot_observable(msgs2)

        def create():
            return reactivex.merge(o1, o2)

        results = scheduler.start(create).messages
        assert len(results) == 9
        for i, result in enumerate(results[:-1]):
            assert result.value.kind == "N"
            assert result.time == 210 + i * 5
            assert result.value.value == i + 2

        assert results[8].value.kind == "C" and results[8].time == 250

    def test_merge_lots3(self):
        scheduler = TestScheduler()
        msgs1 = [
            on_next(150, 1),
            on_next(210, 2),
            on_next(225, 5),
            on_next(240, 8),
            on_completed(245),
        ]
        msgs2 = [
            on_next(150, 1),
            on_next(215, 3),
            on_next(230, 6),
            on_next(245, 9),
            on_completed(250),
        ]
        msgs3 = [on_next(150, 1), on_next(220, 4), on_next(235, 7), on_completed(240)]
        o1 = scheduler.create_hot_observable(msgs1)
        o2 = scheduler.create_hot_observable(msgs2)
        o3 = scheduler.create_hot_observable(msgs3)

        def create():
            return reactivex.merge(o1, o2, o3)

        results = scheduler.start(create).messages
        assert len(results) == 9
        for i, result in enumerate(results[:-1]):
            assert (
                results[i].value.kind == "N"
                and results[i].time == 210 + i * 5
                and results[i].value.value == i + 2
            )

        assert results[8].value.kind == "C" and results[8].time == 250

    def test_merge_error_left(self):
        ex = "ex"
        scheduler = TestScheduler()
        msgs1 = [on_next(150, 1), on_next(210, 2), on_error(245, ex)]
        msgs2 = [on_next(150, 1), on_next(215, 3), on_completed(250)]
        o1 = scheduler.create_hot_observable(msgs1)
        o2 = scheduler.create_hot_observable(msgs2)

        def create():
            return reactivex.merge(o1, o2)

        results = scheduler.start(create)
        assert results.messages == [on_next(210, 2), on_next(215, 3), on_error(245, ex)]

    def test_merge_error_causes_disposal(self):
        ex = "ex"
        scheduler = TestScheduler()
        msgs1 = [on_next(150, 1), on_error(210, ex)]
        msgs2 = [on_next(150, 1), on_next(220, 1), on_completed(250)]
        source_not_disposed = [False]
        o1 = scheduler.create_hot_observable(msgs1)

        def action():
            source_not_disposed[0] = True

        o2 = scheduler.create_hot_observable(msgs2).pipe(ops.do_action(on_next=action))

        def create():
            return reactivex.merge(o1, o2)

        results = scheduler.start(create)

        assert results.messages == [on_error(210, ex)]
        assert not source_not_disposed[0]

    def test_merge_observable_of_observable_data(self):
        scheduler = TestScheduler()
        xs = scheduler.create_hot_observable(
            on_next(
                300,
                scheduler.create_cold_observable(
                    on_next(10, 101),
                    on_next(20, 102),
                    on_next(110, 103),
                    on_next(120, 104),
                    on_next(210, 105),
                    on_next(220, 106),
                    on_completed(230),
                ),
            ),
            on_next(
                400,
                scheduler.create_cold_observable(
                    on_next(10, 201),
                    on_next(20, 202),
                    on_next(30, 203),
                    on_next(40, 200),
                    on_completed(50),
                ),
            ),
            on_next(
                500,
                scheduler.create_cold_observable(
                    on_next(10, 301),
                    on_next(20, 302),
                    on_next(30, 303),
                    on_next(40, 304),
                    on_next(120, 305),
                    on_completed(150),
                ),
            ),
            on_completed(600),
        )

        def create():
            return xs.pipe(ops.merge_all())

        results = scheduler.start(create)
        assert results.messages == [
            on_next(310, 101),
            on_next(320, 102),
            on_next(410, 103),
            on_next(410, 201),
            on_next(420, 104),
            on_next(420, 202),
            on_next(430, 203),
            on_next(440, 200),
            on_next(510, 105),
            on_next(510, 301),
            on_next(520, 106),
            on_next(520, 302),
            on_next(530, 303),
            on_next(540, 304),
            on_next(620, 305),
            on_completed(650),
        ]

    def test_merge_observable_of_observable_data_non_overlapped(self):
        scheduler = TestScheduler()
        xs = scheduler.create_hot_observable(
            on_next(
                300,
                scheduler.create_cold_observable(
                    on_next(10, 101), on_next(20, 102), on_completed(230)
                ),
            ),
            on_next(
                400,
                scheduler.create_cold_observable(
                    on_next(10, 201),
                    on_next(20, 202),
                    on_next(30, 203),
                    on_next(40, 200),
                    on_completed(50),
                ),
            ),
            on_next(
                500,
                scheduler.create_cold_observable(
                    on_next(10, 301),
                    on_next(20, 302),
                    on_next(30, 303),
                    on_next(40, 304),
                    on_completed(50),
                ),
            ),
            on_completed(600),
        )

        def create():
            return xs.pipe(ops.merge_all())

        results = scheduler.start(create)
        assert results.messages == [
            on_next(310, 101),
            on_next(320, 102),
            on_next(410, 201),
            on_next(420, 202),
            on_next(430, 203),
            on_next(440, 200),
            on_next(510, 301),
            on_next(520, 302),
            on_next(530, 303),
            on_next(540, 304),
            on_completed(600),
        ]

    def test_merge_observable_of_observable_inner_throws(self):
        ex = "ex"
        scheduler = TestScheduler()
        xs = scheduler.create_hot_observable(
            on_next(
                300,
                scheduler.create_cold_observable(
                    on_next(10, 101), on_next(20, 102), on_completed(230)
                ),
            ),
            on_next(
                400,
                scheduler.create_cold_observable(
                    on_next(10, 201),
                    on_next(20, 202),
                    on_next(30, 203),
                    on_next(40, 200),
                    on_error(50, ex),
                ),
            ),
            on_next(
                500,
                scheduler.create_cold_observable(
                    on_next(10, 301),
                    on_next(20, 302),
                    on_next(30, 303),
                    on_next(40, 304),
                    on_completed(50),
                ),
            ),
            on_completed(600),
        )

        def create():
            return xs.pipe(ops.merge_all())

        results = scheduler.start(create)
        assert results.messages == [
            on_next(310, 101),
            on_next(320, 102),
            on_next(410, 201),
            on_next(420, 202),
            on_next(430, 203),
            on_next(440, 200),
            on_error(450, ex),
        ]

    def test_merge_observable_of_observable_outer_throws(self):
        ex = "ex"
        scheduler = TestScheduler()
        xs = scheduler.create_hot_observable(
            on_next(
                300,
                scheduler.create_cold_observable(
                    on_next(10, 101), on_next(20, 102), on_completed(230)
                ),
            ),
            on_next(
                400,
                scheduler.create_cold_observable(
                    on_next(10, 201),
                    on_next(20, 202),
                    on_next(30, 203),
                    on_next(40, 200),
                    on_completed(50),
                ),
            ),
            on_error(500, ex),
        )

        def create():
            return xs.pipe(ops.merge_all())

        results = scheduler.start(create)
        assert results.messages == [
            on_next(310, 101),
            on_next(320, 102),
            on_next(410, 201),
            on_next(420, 202),
            on_next(430, 203),
            on_next(440, 200),
            on_error(500, ex),
        ]

    def test_mergeconcat_basic(self):
        scheduler = TestScheduler()
        xs = scheduler.create_hot_observable(
            on_next(
                210,
                scheduler.create_cold_observable(
                    on_next(50, 1), on_next(100, 2), on_next(120, 3), on_completed(140)
                ),
            ),
            on_next(
                260,
                scheduler.create_cold_observable(
                    on_next(20, 4), on_next(70, 5), on_completed(200)
                ),
            ),
            on_next(
                270,
                scheduler.create_cold_observable(
                    on_next(10, 6), on_next(90, 7), on_next(110, 8), on_completed(130)
                ),
            ),
            on_next(
                320,
                scheduler.create_cold_observable(
                    on_next(210, 9), on_next(240, 10), on_completed(300)
                ),
            ),
            on_completed(400),
        )

        def create():
            return xs.pipe(ops.merge(max_concurrent=2))

        results = scheduler.start(create)

        assert results.messages == [
            on_next(260, 1),
            on_next(280, 4),
            on_next(310, 2),
            on_next(330, 3),
            on_next(330, 5),
            on_next(360, 6),
            on_next(440, 7),
            on_next(460, 8),
            on_next(670, 9),
            on_next(700, 10),
            on_completed(760),
        ]
        assert xs.subscriptions == [subscribe(200, 400)]

    def test_mergeconcat_basic_long(self):
        scheduler = TestScheduler()
        xs = scheduler.create_hot_observable(
            on_next(
                210,
                scheduler.create_cold_observable(
                    on_next(50, 1), on_next(100, 2), on_next(120, 3), on_completed(140)
                ),
            ),
            on_next(
                260,
                scheduler.create_cold_observable(
                    on_next(20, 4), on_next(70, 5), on_completed(300)
                ),
            ),
            on_next(
                270,
                scheduler.create_cold_observable(
                    on_next(10, 6), on_next(90, 7), on_next(110, 8), on_completed(130)
                ),
            ),
            on_next(
                320,
                scheduler.create_cold_observable(
                    on_next(210, 9), on_next(240, 10), on_completed(300)
                ),
            ),
            on_completed(400),
        )

        def create():
            return xs.pipe(ops.merge(max_concurrent=2))

        results = scheduler.start(create)

        assert results.messages == [
            on_next(260, 1),
            on_next(280, 4),
            on_next(310, 2),
            on_next(330, 3),
            on_next(330, 5),
            on_next(360, 6),
            on_next(440, 7),
            on_next(460, 8),
            on_next(690, 9),
            on_next(720, 10),
            on_completed(780),
        ]
        assert xs.subscriptions == [subscribe(200, 400)]

    def test_mergeconcat_basic_wide(self):
        scheduler = TestScheduler()
        xs = scheduler.create_hot_observable(
            on_next(
                210,
                scheduler.create_cold_observable(
                    on_next(50, 1), on_next(100, 2), on_next(120, 3), on_completed(140)
                ),
            ),
            on_next(
                260,
                scheduler.create_cold_observable(
                    on_next(20, 4), on_next(70, 5), on_completed(300)
                ),
            ),
            on_next(
                270,
                scheduler.create_cold_observable(
                    on_next(10, 6), on_next(90, 7), on_next(110, 8), on_completed(130)
                ),
            ),
            on_next(
                420,
                scheduler.create_cold_observable(
                    on_next(210, 9), on_next(240, 10), on_completed(300)
                ),
            ),
            on_completed(450),
        )

        def create():
            return xs.pipe(ops.merge(max_concurrent=3))

        results = scheduler.start(create)

        assert results.messages == [
            on_next(260, 1),
            on_next(280, 4),
            on_next(280, 6),
            on_next(310, 2),
            on_next(330, 3),
            on_next(330, 5),
            on_next(360, 7),
            on_next(380, 8),
            on_next(630, 9),
            on_next(660, 10),
            on_completed(720),
        ]
        assert xs.subscriptions == [subscribe(200, 450)]

    def test_mergeconcat_basic_late(self):
        scheduler = TestScheduler()
        xs = scheduler.create_hot_observable(
            on_next(
                210,
                scheduler.create_cold_observable(
                    on_next(50, 1), on_next(100, 2), on_next(120, 3), on_completed(140)
                ),
            ),
            on_next(
                260,
                scheduler.create_cold_observable(
                    on_next(20, 4), on_next(70, 5), on_completed(300)
                ),
            ),
            on_next(
                270,
                scheduler.create_cold_observable(
                    on_next(10, 6), on_next(90, 7), on_next(110, 8), on_completed(130)
                ),
            ),
            on_next(
                420,
                scheduler.create_cold_observable(
                    on_next(210, 9), on_next(240, 10), on_completed(300)
                ),
            ),
            on_completed(750),
        )

        def create():
            return xs.pipe(ops.merge(max_concurrent=3))

        results = scheduler.start(create)

        assert results.messages == [
            on_next(260, 1),
            on_next(280, 4),
            on_next(280, 6),
            on_next(310, 2),
            on_next(330, 3),
            on_next(330, 5),
            on_next(360, 7),
            on_next(380, 8),
            on_next(630, 9),
            on_next(660, 10),
            on_completed(750),
        ]
        assert xs.subscriptions == [subscribe(200, 750)]

    def test_mergeconcat_disposed(self):
        scheduler = TestScheduler()
        xs = scheduler.create_hot_observable(
            on_next(
                210,
                scheduler.create_cold_observable(
                    on_next(50, 1), on_next(100, 2), on_next(120, 3), on_completed(140)
                ),
            ),
            on_next(
                260,
                scheduler.create_cold_observable(
                    on_next(20, 4), on_next(70, 5), on_completed(200)
                ),
            ),
            on_next(
                270,
                scheduler.create_cold_observable(
                    on_next(10, 6), on_next(90, 7), on_next(110, 8), on_completed(130)
                ),
            ),
            on_next(
                320,
                scheduler.create_cold_observable(
                    on_next(210, 9), on_next(240, 10), on_completed(300)
                ),
            ),
            on_completed(400),
        )

        def create():
            return xs.pipe(ops.merge(max_concurrent=2))

        results = scheduler.start(create, disposed=450)
        assert results.messages == [
            on_next(260, 1),
            on_next(280, 4),
            on_next(310, 2),
            on_next(330, 3),
            on_next(330, 5),
            on_next(360, 6),
            on_next(440, 7),
        ]
        assert xs.subscriptions == [subscribe(200, 400)]

    def test_mergeconcat_outererror(self):
        ex = "ex"
        scheduler = TestScheduler()
        xs = scheduler.create_hot_observable(
            on_next(
                210,
                scheduler.create_cold_observable(
                    on_next(50, 1), on_next(100, 2), on_next(120, 3), on_completed(140)
                ),
            ),
            on_next(
                260,
                scheduler.create_cold_observable(
                    on_next(20, 4), on_next(70, 5), on_completed(200)
                ),
            ),
            on_next(
                270,
                scheduler.create_cold_observable(
                    on_next(10, 6), on_next(90, 7), on_next(110, 8), on_completed(130)
                ),
            ),
            on_next(
                320,
                scheduler.create_cold_observable(
                    on_next(210, 9), on_next(240, 10), on_completed(300)
                ),
            ),
            on_error(400, ex),
        )

        def create():
            return xs.pipe(ops.merge(max_concurrent=2))

        results = scheduler.start(create)

        assert results.messages == [
            on_next(260, 1),
            on_next(280, 4),
            on_next(310, 2),
            on_next(330, 3),
            on_next(330, 5),
            on_next(360, 6),
            on_error(400, ex),
        ]
        assert xs.subscriptions == [subscribe(200, 400)]

    def test_mergeconcat_innererror(self):
        ex = "ex"
        scheduler = TestScheduler()
        xs = scheduler.create_hot_observable(
            on_next(
                210,
                scheduler.create_cold_observable(
                    on_next(50, 1), on_next(100, 2), on_next(120, 3), on_completed(140)
                ),
            ),
            on_next(
                260,
                scheduler.create_cold_observable(
                    on_next(20, 4), on_next(70, 5), on_completed(200)
                ),
            ),
            on_next(
                270,
                scheduler.create_cold_observable(
                    on_next(10, 6), on_next(90, 7), on_next(110, 8), on_error(140, ex)
                ),
            ),
            on_next(
                320,
                scheduler.create_cold_observable(
                    on_next(210, 9), on_next(240, 10), on_completed(300)
                ),
            ),
            on_completed(400),
        )

        def create():
            return xs.pipe(ops.merge(max_concurrent=2))

        results = scheduler.start(create)

        assert results.messages == [
            on_next(260, 1),
            on_next(280, 4),
            on_next(310, 2),
            on_next(330, 3),
            on_next(330, 5),
            on_next(360, 6),
            on_next(440, 7),
            on_next(460, 8),
            on_error(490, ex),
        ]
        assert xs.subscriptions == [subscribe(200, 400)]

    def test_merge_112233(self):
        scheduler = TestScheduler()
        xs = scheduler.create_hot_observable(
            on_next(250, 1), on_next(300, 2), on_next(350, 3), on_completed(360)
        )
        ys = scheduler.create_hot_observable(
            on_next(250, 1), on_next(300, 2), on_next(320, 3), on_completed(340)
        )

        def create():
            return xs.pipe(ops.merge(ys))

        results = scheduler.start(create)

        assert results.messages == [
            on_next(250, 1),
            on_next(250, 1),
            on_next(300, 2),
            on_next(300, 2),
            on_next(320, 3),
            on_next(350, 3),
            on_completed(360),
        ]
        assert xs.subscriptions == [subscribe(200, 360)]



import asyncio

import reactivex

stream = reactivex.just("Hello, world!")


async def hello_world():
    n = await stream
    print(n)


loop = asyncio.get_event_loop()
# Blocking call which returns when the hello_world() coroutine is done
loop.run_until_complete(hello_world())
loop.close()



import unittest

from reactivex import create, empty, return_value, throw
from reactivex.disposable import SerialDisposable
from reactivex.operators import map, map_indexed
from reactivex.testing import ReactiveTest, TestScheduler

on_next = ReactiveTest.on_next
on_completed = ReactiveTest.on_completed
on_error = ReactiveTest.on_error
subscribe = ReactiveTest.subscribe
subscribed = ReactiveTest.subscribed
disposed = ReactiveTest.disposed
created = ReactiveTest.created


class RxException(Exception):
    pass


# Helper function for raising exceptions within lambdas


def _raise(ex):
    raise RxException(ex)


class TestSelect(unittest.TestCase):
    def test_map_throws(self):
        mapper = map(lambda x: x)
        with self.assertRaises(RxException):
            return_value(1).pipe(mapper).subscribe(lambda x: _raise("ex"))

        with self.assertRaises(RxException):
            throw("ex").pipe(mapper).subscribe(on_error=lambda ex: _raise(ex))

        with self.assertRaises(RxException):
            empty().pipe(mapper).subscribe(
                lambda x: x, lambda ex: ex, lambda: _raise("ex")
            )

        def subscribe(observer, scheduler=None):
            _raise("ex")

        with self.assertRaises(RxException):
            create(subscribe).pipe(map(lambda x: x)).subscribe()

    def test_map_disposeinsidemapper(self):
        scheduler = TestScheduler()
        xs = scheduler.create_hot_observable(
            on_next(100, 1), on_next(200, 2), on_next(500, 3), on_next(600, 4)
        )
        results = scheduler.create_observer()
        d = SerialDisposable()
        invoked = [0]

        def projection(x, *args, **kw):
            invoked[0] += 1

            if scheduler.clock > 400:
                d.dispose()
            return x

        d.disposable = xs.pipe(map(projection)).subscribe(results, scheduler)

        def action(scheduler, state):
            return d.dispose()

        scheduler.schedule_absolute(ReactiveTest.disposed, action)
        scheduler.start()

        assert results.messages == [on_next(100, 1), on_next(200, 2)]
        assert xs.subscriptions == [ReactiveTest.subscribe(0, 500)]

        assert invoked[0] == 3

    def test_map_completed(self):
        scheduler = TestScheduler()
        xs = scheduler.create_hot_observable(
            on_next(180, 1),
            on_next(210, 2),
            on_next(240, 3),
            on_next(290, 4),
            on_next(350, 5),
            on_completed(400),
            on_next(410, -1),
            on_completed(420),
            on_error(430, "ex"),
        )
        invoked = [0]

        def factory():
            def projection(x):
                invoked[0] += 1
                return x + 1

            return xs.pipe(map(projection))

        results = scheduler.start(factory)
        assert results.messages == [
            on_next(210, 3),
            on_next(240, 4),
            on_next(290, 5),
            on_next(350, 6),
            on_completed(400),
        ]
        assert xs.subscriptions == [ReactiveTest.subscribe(200, 400)]
        assert invoked[0] == 4

    def test_map_default_mapper(self):
        scheduler = TestScheduler()
        xs = scheduler.create_hot_observable(
            on_next(180, 1),
            on_next(210, 2),
            on_next(240, 3),
            on_next(290, 4),
            on_next(350, 5),
            on_completed(400),
            on_next(410, -1),
            on_completed(420),
            on_error(430, "ex"),
        )

        def factory():
            return xs.pipe(map())

        results = scheduler.start(factory)
        assert results.messages == [
            on_next(210, 2),
            on_next(240, 3),
            on_next(290, 4),
            on_next(350, 5),
            on_completed(400),
        ]
        assert xs.subscriptions == [ReactiveTest.subscribe(200, 400)]

    def test_map_completed_two(self):
        for i in range(100):
            scheduler = TestScheduler()
            invoked = [0]

            xs = scheduler.create_hot_observable(
                on_next(180, 1),
                on_next(210, 2),
                on_next(240, 3),
                on_next(290, 4),
                on_next(350, 5),
                on_completed(400),
                on_next(410, -1),
                on_completed(420),
                on_error(430, "ex"),
            )

            def factory():
                def projection(x):
                    invoked[0] += 1
                    return x + 1

                return xs.pipe(map(projection))

            results = scheduler.start(factory)
            assert results.messages == [
                on_next(210, 3),
                on_next(240, 4),
                on_next(290, 5),
                on_next(350, 6),
                on_completed(400),
            ]
            assert xs.subscriptions == [subscribe(200, 400)]
            assert invoked[0] == 4

    def test_map_not_completed(self):
        scheduler = TestScheduler()
        invoked = [0]
        xs = scheduler.create_hot_observable(
            on_next(180, 1),
            on_next(210, 2),
            on_next(240, 3),
            on_next(290, 4),
            on_next(350, 5),
        )

        def factory():
            def projection(x):
                invoked[0] += 1
                return x + 1

            return xs.pipe(map(projection))

        results = scheduler.start(factory)
        assert results.messages == [
            on_next(210, 3),
            on_next(240, 4),
            on_next(290, 5),
            on_next(350, 6),
        ]
        assert xs.subscriptions == [subscribe(200, 1000)]
        assert invoked[0] == 4

    def test_map_error(self):
        scheduler = TestScheduler()
        ex = "ex"
        invoked = [0]
        xs = scheduler.create_hot_observable(
            on_next(180, 1),
            on_next(210, 2),
            on_next(240, 3),
            on_next(290, 4),
            on_next(350, 5),
            on_error(400, ex),
            on_next(410, -1),
            on_completed(420),
            on_error(430, "ex"),
        )

        def factory():
            def projection(x):
                invoked[0] += 1
                return x + 1

            return xs.pipe(map(projection))

        results = scheduler.start(factory)
        assert results.messages == [
            on_next(210, 3),
            on_next(240, 4),
            on_next(290, 5),
            on_next(350, 6),
            on_error(400, ex),
        ]
        assert xs.subscriptions == [subscribe(200, 400)]
        assert invoked[0] == 4

    def test_map_mapper_throws(self):
        scheduler = TestScheduler()
        invoked = [0]
        ex = "ex"
        xs = scheduler.create_hot_observable(
            on_next(180, 1),
            on_next(210, 2),
            on_next(240, 3),
            on_next(290, 4),
            on_next(350, 5),
            on_completed(400),
            on_next(410, -1),
            on_completed(420),
            on_error(430, "ex"),
        )

        def factory():
            def projection(x):
                invoked[0] += 1
                if invoked[0] == 3:
                    raise Exception(ex)

                return x + 1

            return xs.pipe(map(projection))

        results = scheduler.start(factory)
        assert results.messages == [on_next(210, 3), on_next(240, 4), on_error(290, ex)]
        assert xs.subscriptions == [subscribe(200, 290)]
        assert invoked[0] == 3

    def test_map_with_index_throws(self):
        with self.assertRaises(RxException):
            mapper = map_indexed(lambda x, index: x)

            return return_value(1).pipe(mapper).subscribe(lambda x: _raise("ex"))

        with self.assertRaises(RxException):
            return (
                throw("ex").pipe(mapper).subscribe(lambda x: x, lambda ex: _raise(ex))
            )

        with self.assertRaises(RxException):
            return (
                empty()
                .pipe(mapper)
                .subscribe(lambda x: x, lambda ex: None, lambda: _raise("ex"))
            )

        with self.assertRaises(RxException):
            return create(lambda o, s: _raise("ex")).pipe(mapper).subscribe()

    def test_map_with_index_dispose_inside_mapper(self):
        scheduler = TestScheduler()
        xs = scheduler.create_hot_observable(
            on_next(100, 4), on_next(200, 3), on_next(500, 2), on_next(600, 1)
        )
        invoked = [0]
        results = scheduler.create_observer()
        d = SerialDisposable()

        def projection(x, index):
            invoked[0] += 1
            if scheduler.clock > 400:
                d.dispose()

            return x + index * 10

        d.disposable = xs.pipe(map_indexed(projection)).subscribe(results)

        def action(scheduler, state):
            return d.dispose()

        scheduler.schedule_absolute(disposed, action)
        scheduler.start()
        assert results.messages == [on_next(100, 4), on_next(200, 13)]
        assert xs.subscriptions == [subscribe(0, 500)]
        assert invoked[0] == 3

    def test_map_with_index_completed(self):
        scheduler = TestScheduler()
        invoked = [0]
        xs = scheduler.create_hot_observable(
            on_next(180, 5),
            on_next(210, 4),
            on_next(240, 3),
            on_next(290, 2),
            on_next(350, 1),
            on_completed(400),
            on_next(410, -1),
            on_completed(420),
            on_error(430, "ex"),
        )

        def factory():
            def projection(x, index):
                invoked[0] += 1
                return (x + 1) + (index * 10)

            return xs.pipe(map_indexed(projection))

        results = scheduler.start(factory)
        assert results.messages == [
            on_next(210, 5),
            on_next(240, 14),
            on_next(290, 23),
            on_next(350, 32),
            on_completed(400),
        ]
        assert xs.subscriptions == [subscribe(200, 400)]
        assert invoked[0] == 4

    def test_map_with_index_default_mapper(self):
        scheduler = TestScheduler()
        xs = scheduler.create_hot_observable(
            on_next(180, 5),
            on_next(210, 4),
            on_next(240, 3),
            on_next(290, 2),
            on_next(350, 1),
            on_completed(400),
            on_next(410, -1),
            on_completed(420),
            on_error(430, "ex"),
        )

        def factory():
            return xs.pipe(map_indexed())

        results = scheduler.start(factory)
        assert results.messages == [
            on_next(210, 4),
            on_next(240, 3),
            on_next(290, 2),
            on_next(350, 1),
            on_completed(400),
        ]

        assert xs.subscriptions == [subscribe(200, 400)]

    def test_map_with_index_not_completed(self):
        scheduler = TestScheduler()
        invoked = [0]
        xs = scheduler.create_hot_observable(
            on_next(180, 5),
            on_next(210, 4),
            on_next(240, 3),
            on_next(290, 2),
            on_next(350, 1),
        )

        def factory():
            def projection(x, index):
                invoked[0] += 1
                return (x + 1) + (index * 10)

            return xs.pipe(map_indexed(projection))

        results = scheduler.start(factory)
        assert results.messages == [
            on_next(210, 5),
            on_next(240, 14),
            on_next(290, 23),
            on_next(350, 32),
        ]
        assert xs.subscriptions == [subscribe(200, 1000)]
        assert invoked[0] == 4

    def test_map_with_index_error(self):
        scheduler = TestScheduler()
        ex = "ex"
        invoked = [0]
        xs = scheduler.create_hot_observable(
            on_next(180, 5),
            on_next(210, 4),
            on_next(240, 3),
            on_next(290, 2),
            on_next(350, 1),
            on_error(400, ex),
            on_next(410, -1),
            on_completed(420),
            on_error(430, "ex"),
        )

        def factory():
            def projection(x, index):
                invoked[0] += 1
                return (x + 1) + (index * 10)

            return xs.pipe(map_indexed(projection))

        results = scheduler.start(factory)

        assert results.messages == [
            on_next(210, 5),
            on_next(240, 14),
            on_next(290, 23),
            on_next(350, 32),
            on_error(400, ex),
        ]
        assert xs.subscriptions == [subscribe(200, 400)]
        assert invoked[0] == 4

    def test_map_with_index_mapper_throws(self):
        scheduler = TestScheduler()
        invoked = [0]
        ex = "ex"
        xs = scheduler.create_hot_observable(
            on_next(180, 5),
            on_next(210, 4),
            on_next(240, 3),
            on_next(290, 2),
            on_next(350, 1),
            on_completed(400),
            on_next(410, -1),
            on_completed(420),
            on_error(430, "ex"),
        )

        def factory():
            def projection(x, index):
                invoked[0] += 1
                if invoked[0] == 3:
                    raise Exception(ex)
                return (x + 1) + (index * 10)

            return xs.pipe(map_indexed(projection))

        results = scheduler.start(factory)
        assert results.messages == [
            on_next(210, 5),
            on_next(240, 14),
            on_error(290, ex),
        ]
        assert xs.subscriptions == [subscribe(200, 290)]
        assert invoked[0] == 3


if __name__ == "__main__":
    unittest.main()



import unittest

from reactivex import operators as ops
from reactivex.testing import ReactiveTest, TestScheduler

on_next = ReactiveTest.on_next
on_completed = ReactiveTest.on_completed
on_error = ReactiveTest.on_error
subscribe = ReactiveTest.subscribe
subscribed = ReactiveTest.subscribed
disposed = ReactiveTest.disposed
created = ReactiveTest.created


class TestSkip(unittest.TestCase):
    def test_skip_complete_after(self):
        scheduler = TestScheduler()
        xs = scheduler.create_hot_observable(
            on_next(70, 6),
            on_next(150, 4),
            on_next(210, 9),
            on_next(230, 13),
            on_next(270, 7),
            on_next(280, 1),
            on_next(300, -1),
            on_next(310, 3),
            on_next(340, 8),
            on_next(370, 11),
            on_next(410, 15),
            on_next(415, 16),
            on_next(460, 72),
            on_next(510, 76),
            on_next(560, 32),
            on_next(570, -100),
            on_next(580, -3),
            on_next(590, 5),
            on_next(630, 10),
            on_completed(690),
        )

        def create():
            return xs.pipe(ops.skip(20))

        results = scheduler.start(create)

        assert results.messages == [on_completed(690)]
        assert xs.subscriptions == [subscribe(200, 690)]

    def test_skip_complete_same(self):
        scheduler = TestScheduler()
        xs = scheduler.create_hot_observable(
            on_next(70, 6),
            on_next(150, 4),
            on_next(210, 9),
            on_next(230, 13),
            on_next(270, 7),
            on_next(280, 1),
            on_next(300, -1),
            on_next(310, 3),
            on_next(340, 8),
            on_next(370, 11),
            on_next(410, 15),
            on_next(415, 16),
            on_next(460, 72),
            on_next(510, 76),
            on_next(560, 32),
            on_next(570, -100),
            on_next(580, -3),
            on_next(590, 5),
            on_next(630, 10),
            on_completed(690),
        )

        def create():
            return xs.pipe(ops.skip(17))

        results = scheduler.start(create)

        assert results.messages == [on_completed(690)]
        assert xs.subscriptions == [subscribe(200, 690)]

    def test_skip_complete_before(self):
        scheduler = TestScheduler()
        xs = scheduler.create_hot_observable(
            on_next(70, 6),
            on_next(150, 4),
            on_next(210, 9),
            on_next(230, 13),
            on_next(270, 7),
            on_next(280, 1),
            on_next(300, -1),
            on_next(310, 3),
            on_next(340, 8),
            on_next(370, 11),
            on_next(410, 15),
            on_next(415, 16),
            on_next(460, 72),
            on_next(510, 76),
            on_next(560, 32),
            on_next(570, -100),
            on_next(580, -3),
            on_next(590, 5),
            on_next(630, 10),
            on_completed(690),
        )

        def create():
            return xs.pipe(ops.skip(10))

        results = scheduler.start(create)

        assert results.messages == [
            on_next(460, 72),
            on_next(510, 76),
            on_next(560, 32),
            on_next(570, -100),
            on_next(580, -3),
            on_next(590, 5),
            on_next(630, 10),
            on_completed(690),
        ]
        assert xs.subscriptions == [subscribe(200, 690)]

    def test_skip_Complete_zero(self):
        scheduler = TestScheduler()
        xs = scheduler.create_hot_observable(
            on_next(70, 6),
            on_next(150, 4),
            on_next(210, 9),
            on_next(230, 13),
            on_next(270, 7),
            on_next(280, 1),
            on_next(300, -1),
            on_next(310, 3),
            on_next(340, 8),
            on_next(370, 11),
            on_next(410, 15),
            on_next(415, 16),
            on_next(460, 72),
            on_next(510, 76),
            on_next(560, 32),
            on_next(570, -100),
            on_next(580, -3),
            on_next(590, 5),
            on_next(630, 10),
            on_completed(690),
        )

        def create():
            return xs.pipe(ops.skip(0))

        results = scheduler.start(create)

        assert results.messages == [
            on_next(210, 9),
            on_next(230, 13),
            on_next(270, 7),
            on_next(280, 1),
            on_next(300, -1),
            on_next(310, 3),
            on_next(340, 8),
            on_next(370, 11),
            on_next(410, 15),
            on_next(415, 16),
            on_next(460, 72),
            on_next(510, 76),
            on_next(560, 32),
            on_next(570, -100),
            on_next(580, -3),
            on_next(590, 5),
            on_next(630, 10),
            on_completed(690),
        ]
        assert xs.subscriptions == [subscribe(200, 690)]

    def test_skip_error_after(self):
        ex = "ex"
        scheduler = TestScheduler()
        xs = scheduler.create_hot_observable(
            on_next(70, 6),
            on_next(150, 4),
            on_next(210, 9),
            on_next(230, 13),
            on_next(270, 7),
            on_next(280, 1),
            on_next(300, -1),
            on_next(310, 3),
            on_next(340, 8),
            on_next(370, 11),
            on_next(410, 15),
            on_next(415, 16),
            on_next(460, 72),
            on_next(510, 76),
            on_next(560, 32),
            on_next(570, -100),
            on_next(580, -3),
            on_next(590, 5),
            on_next(630, 10),
            on_error(690, ex),
        )

        def create():
            return xs.pipe(ops.skip(20))

        results = scheduler.start(create)

        assert results.messages == [on_error(690, ex)]
        assert xs.subscriptions == [subscribe(200, 690)]

    def test_skip_error_same(self):
        ex = "ex"
        scheduler = TestScheduler()
        xs = scheduler.create_hot_observable(
            on_next(70, 6),
            on_next(150, 4),
            on_next(210, 9),
            on_next(230, 13),
            on_next(270, 7),
            on_next(280, 1),
            on_next(300, -1),
            on_next(310, 3),
            on_next(340, 8),
            on_next(370, 11),
            on_next(410, 15),
            on_next(415, 16),
            on_next(460, 72),
            on_next(510, 76),
            on_next(560, 32),
            on_next(570, -100),
            on_next(580, -3),
            on_next(590, 5),
            on_next(630, 10),
            on_error(690, ex),
        )

        def create():
            return xs.pipe(ops.skip(17))

        results = scheduler.start(create)

        assert results.messages == [on_error(690, ex)]
        assert xs.subscriptions == [subscribe(200, 690)]

    def test_skip_error_before(self):
        ex = "ex"
        scheduler = TestScheduler()
        xs = scheduler.create_hot_observable(
            on_next(70, 6),
            on_next(150, 4),
            on_next(210, 9),
            on_next(230, 13),
            on_next(270, 7),
            on_next(280, 1),
            on_next(300, -1),
            on_next(310, 3),
            on_next(340, 8),
            on_next(370, 11),
            on_next(410, 15),
            on_next(415, 16),
            on_next(460, 72),
            on_next(510, 76),
            on_next(560, 32),
            on_next(570, -100),
            on_next(580, -3),
            on_next(590, 5),
            on_next(630, 10),
            on_error(690, ex),
        )

        def create():
            return xs.pipe(ops.skip(3))

        results = scheduler.start(create)

        assert results.messages == [
            on_next(280, 1),
            on_next(300, -1),
            on_next(310, 3),
            on_next(340, 8),
            on_next(370, 11),
            on_next(410, 15),
            on_next(415, 16),
            on_next(460, 72),
            on_next(510, 76),
            on_next(560, 32),
            on_next(570, -100),
            on_next(580, -3),
            on_next(590, 5),
            on_next(630, 10),
            on_error(690, ex),
        ]
        assert xs.subscriptions == [subscribe(200, 690)]

    def test_skip_dispose_before(self):
        scheduler = TestScheduler()
        xs = scheduler.create_hot_observable(
            on_next(70, 6),
            on_next(150, 4),
            on_next(210, 9),
            on_next(230, 13),
            on_next(270, 7),
            on_next(280, 1),
            on_next(300, -1),
            on_next(310, 3),
            on_next(340, 8),
            on_next(370, 11),
            on_next(410, 15),
            on_next(415, 16),
            on_next(460, 72),
            on_next(510, 76),
            on_next(560, 32),
            on_next(570, -100),
            on_next(580, -3),
            on_next(590, 5),
            on_next(630, 10),
        )

        def create():
            return xs.pipe(ops.skip(3))

        results = scheduler.start(create, disposed=250)
        assert results.messages == []
        assert xs.subscriptions == [subscribe(200, 250)]

    def test_skip_dispose_after(self):
        scheduler = TestScheduler()
        xs = scheduler.create_hot_observable(
            on_next(70, 6),
            on_next(150, 4),
            on_next(210, 9),
            on_next(230, 13),
            on_next(270, 7),
            on_next(280, 1),
            on_next(300, -1),
            on_next(310, 3),
            on_next(340, 8),
            on_next(370, 11),
            on_next(410, 15),
            on_next(415, 16),
            on_next(460, 72),
            on_next(510, 76),
            on_next(560, 32),
            on_next(570, -100),
            on_next(580, -3),
            on_next(590, 5),
            on_next(630, 10),
        )

        def create():
            return xs.pipe(ops.skip(3))

        results = scheduler.start(create, disposed=400)
        assert results.messages == [
            on_next(280, 1),
            on_next(300, -1),
            on_next(310, 3),
            on_next(340, 8),
            on_next(370, 11),
        ]
        assert xs.subscriptions == [subscribe(200, 400)]


if __name__ == "__main__":
    unittest.main()



from reactivex import operators as ops
from reactivex.testing.marbles import marbles_testing

"""
Tests debounceTime from reactivexjs
https://github.com/ReactiveX/rxjs/blob/master/spec/operators/debounceTime-spec.ts

it should delay all element by the specified time
"""
with marbles_testing(timespan=1.0) as (start, cold, hot, exp):

    e1 = cold("-a--------b------c----|")
    ex = exp("------a--------b------(c,|)")
    expected = ex

    def create():
        return e1.pipe(
            ops.debounce(5),
        )

    results = start(create)
    assert results == expected

print("debounce: results vs expected")
for r, e in zip(results, expected):
    print(r, e)



import reactivex
from reactivex import operators as ops

"""
simple example that merges two cold observables.
"""

source0 = reactivex.cold("a-----d---1--------4-|", timespan=0.01)
source1 = reactivex.cold("--b-c-------2---3-|   ", timespan=0.01)

observable = reactivex.merge(source0, source1).pipe(ops.to_iterable())
elements = observable.run()
print("received {}".format(list(elements)))



import unittest

from reactivex import operators as ops
from reactivex.testing import ReactiveTest, TestScheduler

on_next = ReactiveTest.on_next
on_completed = ReactiveTest.on_completed
on_error = ReactiveTest.on_error
subscribe = ReactiveTest.subscribe
subscribed = ReactiveTest.subscribed
disposed = ReactiveTest.disposed
created = ReactiveTest.created


class TestIsEmpty(unittest.TestCase):
    def test_is_empty_empty(self):
        scheduler = TestScheduler()
        xs = scheduler.create_hot_observable(on_next(150, 1), on_completed(250))

        def create():
            return xs.pipe(ops.is_empty())

        res = scheduler.start(create=create).messages
        assert res == [on_next(250, True), on_completed(250)]
        assert xs.subscriptions == [subscribe(200, 250)]

    def test_is_empty_return(self):
        scheduler = TestScheduler()
        xs = scheduler.create_hot_observable(
            on_next(150, 1), on_next(210, 2), on_completed(250)
        )

        def create():
            return xs.pipe(ops.is_empty())

        res = scheduler.start(create=create).messages
        assert res == [on_next(210, False), on_completed(210)]
        assert xs.subscriptions == [subscribe(200, 210)]

    def test_is_empty_on_error(self):
        ex = "ex"
        scheduler = TestScheduler()
        xs = scheduler.create_hot_observable(on_next(150, 1), on_error(210, ex))

        def create():
            return xs.pipe(ops.is_empty())

        res = scheduler.start(create=create).messages
        assert res == [on_error(210, ex)]
        assert xs.subscriptions == [subscribe(200, 210)]

    def test_is_empty_never(self):
        scheduler = TestScheduler()
        xs = scheduler.create_hot_observable(on_next(150, 1))

        def create():
            return xs.pipe(ops.is_empty())

        res = scheduler.start(create=create).messages
        assert res == []
        assert xs.subscriptions == [subscribe(200, 1000)]



import asyncio
import unittest

from reactivex import operators as ops
from reactivex.scheduler.eventloop import AsyncIOScheduler
from reactivex.subject import Subject


class TestFlatMapAsync(unittest.TestCase):
    def test_flat_map_async(self):
        actual_next = None
        loop = asyncio.get_event_loop()
        scheduler = AsyncIOScheduler(loop=loop)

        def mapper(i: int):
            async def _mapper(i: int):
                return i + 1

            return asyncio.ensure_future(_mapper(i))

        def on_next(i: int):
            nonlocal actual_next
            actual_next = i

        def on_error(ex):
            print("Error", ex)

        async def test_flat_map():
            x: Subject[int] = Subject()
            x.pipe(ops.flat_map(mapper)).subscribe(
                on_next, on_error, scheduler=scheduler
            )
            x.on_next(10)
            await asyncio.sleep(0.1)

        loop.run_until_complete(test_flat_map())
        assert actual_next == 11



import unittest

from reactivex import operators as ops
from reactivex.testing.marbles import marbles_testing


class TestConcatIntegration(unittest.TestCase):
    def test_concat_repeat(self):
        with marbles_testing() as (start, cold, hot, exp):
            e1 = cold("-e11-e12|")
            e2 = cold("-e21-e22|")
            ex = exp("-e11-e12-e21-e22-e11-e12-e21-e22|")

            obs = e1.pipe(ops.concat(e2), ops.repeat(2))

            results = start(obs)
            assert results == ex



import os
import threading
import unittest
from datetime import timedelta
from time import sleep

import pytest

from reactivex.internal.basic import default_now
from reactivex.scheduler import NewThreadScheduler

CI = os.getenv("CI") is not None


class TestNewThreadScheduler(unittest.TestCase):
    def test_new_thread_now(self):
        scheduler = NewThreadScheduler()
        diff = scheduler.now - default_now()
        assert abs(diff) < timedelta(milliseconds=5)

    @pytest.mark.skipif(CI, reason="Flaky test in GitHub Actions")
    def test_new_thread_now_units(self):
        scheduler = NewThreadScheduler()
        diff = scheduler.now
        sleep(1.1)
        diff = scheduler.now - diff
        assert timedelta(milliseconds=1000) < diff < timedelta(milliseconds=1300)

    def test_new_thread_schedule_action(self):
        scheduler = NewThreadScheduler()
        ran = False

        def action(scheduler, state):
            nonlocal ran
            ran = True

        scheduler.schedule(action)

        sleep(0.1)
        assert ran is True

    def test_new_thread_schedule_action_due(self):
        scheduler = NewThreadScheduler()
        starttime = default_now()
        endtime = None

        def action(scheduler, state):
            nonlocal endtime
            endtime = default_now()

        scheduler.schedule_relative(timedelta(milliseconds=200), action)

        sleep(0.4)
        assert endtime is not None
        diff = endtime - starttime
        assert diff > timedelta(milliseconds=180)

    def test_new_thread_schedule_action_cancel(self):
        ran = False
        scheduler = NewThreadScheduler()

        def action(scheduler, state):
            nonlocal ran
            ran = True

        d = scheduler.schedule_relative(timedelta(milliseconds=1), action)
        d.dispose()

        sleep(0.2)
        assert ran is False

    def test_new_thread_schedule_periodic(self):
        scheduler = NewThreadScheduler()
        gate = threading.Semaphore(0)
        period = 0.05
        counter = 3

        def action(state: int):
            nonlocal counter
            if state:
                counter -= 1
                return state - 1
            if counter == 0:
                gate.release()

        scheduler.schedule_periodic(period, action, counter)
        gate.acquire()
        assert counter == 0

    def test_new_thread_schedule_periodic_cancel(self):
        scheduler = NewThreadScheduler()
        period = 0.1
        counter = 4

        def action(state: int):
            nonlocal counter
            if state:
                counter -= 1
                return state - 1

        disp = scheduler.schedule_periodic(period, action, counter)
        sleep(0.4)
        disp.dispose()
        assert 0 <= counter < 4



import tkinter
from tkinter import Event, Frame, Label, Tk
from typing import Any, Tuple

import reactivex
from reactivex import Observable
from reactivex import operators as ops
from reactivex.scheduler.mainloop import TkinterScheduler
from reactivex.subject import Subject


def main() -> None:
    root = Tk()
    root.title("Rx for Python rocks")
    scheduler = TkinterScheduler(root)

    mousemoves: Subject[Event[Any]] = Subject()

    frame = Frame(root, width=600, height=600)
    frame.bind("<Motion>", mousemoves.on_next)

    text = "TIME FLIES LIKE AN ARROW"

    def on_next(info: Tuple[tkinter.Label, "Event[Frame]", int]) -> None:
        label, ev, i = info
        label.place(x=ev.x + i * 12 + 15, y=ev.y)

    def label2stream(
        label: tkinter.Label, index: int
    ) -> Observable[Tuple[tkinter.Label, "Event[Frame]", int]]:

        return mousemoves.pipe(
            ops.map(lambda ev: (label, ev, index)),
            ops.delay(index * 0.1),
        )

    def char2label(char: str) -> Label:
        return Label(frame, text=char, borderwidth=0, padx=0, pady=0)

    reactivex.from_(text).pipe(
        ops.map(char2label),
        ops.flat_map_indexed(label2stream),
    ).subscribe(on_next, on_error=print, scheduler=scheduler)

    frame.pack()
    root.mainloop()


if __name__ == "__main__":
    main()
